from typing import TYPE_CHECKING, Iterator

import kubernetes

if TYPE_CHECKING:
    from .models import PykormModel, NamespacedModel, ClusterModel



def _custom_objects_api():
    return kubernetes.client.CustomObjectsApi()


def _coreV1_api():
    return kubernetes.client.CoreV1Api()


class BaseQuery:
    baseobject: 'PykormModel'

    def __init__(self, baseobject: 'PykormModel'):
        self.baseobject = baseobject

    def filter_by(self, **kwargs: str) -> Iterator['PykormModel']:
        for el in self._iter():
            if el._matches_attributes(kwargs):
                yield el


    def all(self) -> Iterator['PykormModel']:
        for el in self._iter():
            yield el



class NamespacedObjectQuery(BaseQuery):
    def _iter(self) -> Iterator['NamespacedModel']:
        api = _custom_objects_api()
        corev1 = _coreV1_api()
        base_cls = self.baseobject


        for namespace in corev1.list_namespace().items:
            ns_name = namespace.metadata.name
            objs = api.list_namespaced_custom_object(base_cls._pykorm_group, base_cls._pykorm_version, ns_name, base_cls._pykorm_plural)
            for obj in objs['items']:
                yield self.baseobject._instantiate_with_dict(obj)


    def _save(self, obj: 'NamespacedModel'):
        api = _custom_objects_api()

        k8s_dict = obj._k8s_dict

        if obj._k8s_uid is None:
            result = api.create_namespaced_custom_object(obj._pykorm_group, obj._pykorm_version, obj.namespace, obj._pykorm_plural, k8s_dict)
        else:
            result = api.patch_namespaced_custom_object(obj._pykorm_group, obj._pykorm_version, obj.namespace, obj._pykorm_plural, obj.name, k8s_dict)

        obj._set_attributes_with_dict(result)


    def _delete(self, obj: 'NamespacedModel'):
        api = _custom_objects_api()
        api.delete_namespaced_custom_object(obj._pykorm_group, obj._pykorm_version, obj.namespace, obj._pykorm_plural, obj.name)



class ClusterObjectQuery(BaseQuery):
    def _iter(self) -> Iterator['ClusterModel']:
        api = _custom_objects_api()
        base_cls = self.baseobject

        objs = api.list_cluster_custom_object(base_cls._pykorm_group, base_cls._pykorm_version, base_cls._pykorm_plural)
        for obj in objs['items']:
            yield self.baseobject._instantiate_with_dict(obj)


    def _save(self, obj: 'ClusterModel'):
        api = _custom_objects_api()

        k8s_dict = obj._k8s_dict

        if obj._k8s_uid is None:
            result = api.create_cluster_custom_object(obj._pykorm_group, obj._pykorm_version, obj._pykorm_plural, k8s_dict)
        else:
            result = api.patch_cluster_custom_object(obj._pykorm_group, obj._pykorm_version, obj._pykorm_plural, obj.name, k8s_dict)

        obj._set_attributes_with_dict(result)


    def _delete(self, obj: 'ClusterModel'):
        api = _custom_objects_api()
        api.delete_cluster_custom_object(obj._pykorm_group, obj._pykorm_version, obj._pykorm_plural, obj.name)





import inspect
from typing import List, Tuple, Dict

from . import fields
from . import query as pykorm_query


def dict_deep_merge(source: Dict, extra: Dict) -> Dict:
    """
    deep merges two dicts together
    """
    for key, value in source.items():
        if isinstance(value, dict):
            # get node or create one
            node = extra.setdefault(key, {})
            dict_deep_merge(value, node)
        else:
            extra[key] = value

    return extra



class PykormModel:
    name: str = fields.Metadata('name', readonly=True)
    _k8s_uid: str = fields.Metadata('uid', readonly=True)

    _pykorm_group: str
    _pykorm_version: str
    _pykorm_plural: str

    query: pykorm_query.BaseQuery

    @classmethod
    def _get_pykorm_attributes(cls) -> List[Tuple[str, fields.DataField]]:
        attributes = inspect.getmembers(cls, lambda a: not(inspect.isroutine(a)))
        obj_attrs = [a for a in attributes if not(a[0].startswith('__') and a[0].endswith('__'))]

        retl = []
        for obj in obj_attrs:
            (_attr_name, attr) = obj
            if isinstance(attr, fields.DataField):
                retl.append(obj)
        return retl

    def __repr__(self):
        repr_dict = {}
        for (attr_name, _) in self._get_pykorm_attributes():
            repr_dict[attr_name] = getattr(self, attr_name)

        repr_str = [f'{k}={v}' for k, v in repr_dict.items()]

        return f'<{self.__class__.__name__} {" ".join(repr_str)}>'


    def __setattr__(self, item: str, value):
        for (attr_name, attr) in self._get_pykorm_attributes():
            if item == attr_name:
                if attr.readonly and getattr(self, item) is not None:
                    # We allow to set the attribute if it was not set before
                    raise Exception(f'{attr_name} attribute is read_only !')
        self.__dict__[item] = value


    def __getattribute__(self, item: str):
        attr = object.__getattribute__(self, item)

        if isinstance(attr, fields.DataField):
            return None
        else:
            return attr

    def _matches_attributes(self, filters_dict: Dict[str, str]) -> bool:
        for attribute_name, attribute_value in filters_dict.items():
            if getattr(self, attribute_name) != attribute_value:
                return False
        return True


    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False

        return self._k8s_dict == other._k8s_dict


    @classmethod
    def _instantiate_with_dict(cls, k8s_dict) -> 'PykormModel':
        ''' Creates the model with data from the k8s data structure '''
        obj = cls.__new__(cls)
        obj._set_attributes_with_dict(k8s_dict)
        return obj

    def _set_attributes_with_dict(self, k8s_dict: Dict):
        self.__k8s_data = k8s_dict

        for (attr_name, attr_value) in self._get_pykorm_attributes():
            value = attr_value.get_data(k8s_dict)
            self.__dict__[attr_name] = value


    @property
    def _k8s_dict(self):
        '''
        Returns the model as a kubernetes dict/yaml structure
        '''
        d = {
            "apiVersion": f'{self._pykorm_group}/{self._pykorm_version}',
            "kind": self.__class__.__name__,
            "metadata": {
                "name": self.name,
            },
            "spec": {
            }
        }

        for (attr_name, attr_type) in self._get_pykorm_attributes():
            attr_value = getattr(self, attr_name)
            if not isinstance(attr_value, fields.DataField):
                attr_dict_path = attr_type.to_dict(attr_value)
                d = dict_deep_merge(d, attr_dict_path)
        return d


class NamespacedModel(PykormModel):
    namespace = fields.Metadata('namespace')


class ClusterModel(PykormModel):
    pass




from typing import Dict


class DataField:
    _root_dict_key: str
    _split_dots = True
    path: str
    default: object
    readonly: bool = False

    def __init__(self, path: str, default=None, readonly: bool = False):
        self.path = path.strip('.')
        self.default = default
        self.readonly = readonly

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        return f'<pykorm/{cls_name}: {self.path}>'


    def to_dict(self, value: str) -> Dict:
        path = self.path

        leaf = value
        if self._split_dots:
            for dict_item in reversed(path.split('.')):
                leaf = {dict_item: leaf}
        else:
            leaf = {path: value}

        for dict_item in reversed(self._root_dict_key.split('.')):
            leaf = {dict_item: leaf}

        return leaf


    def get_data(self, k8s_dict: Dict):
        for elem in self._root_dict_key.split('.'):
            k8s_dict = k8s_dict.get(elem, {})

        if self._split_dots:
            for elem in self.path.split('.'):
                if elem in k8s_dict:
                    k8s_dict = k8s_dict[elem]
                else:
                    return self.default
        else:
            k8s_dict = k8s_dict.get(self.path, self.default)

        return k8s_dict


class Spec(DataField):
    _root_dict_key = 'spec'


class Metadata(DataField):
    _root_dict_key = 'metadata'


class MetadataAnnotation(DataField):
    _root_dict_key = 'metadata.annotations'
    _split_dots = False





import kubernetes

from . import query as pykorm_query
from . import models



def _process_cls(cls, query_class, group: str, version: str, plural: str):
    cls._pykorm_group = group
    cls._pykorm_version = version
    cls._pykorm_plural = plural
    cls.query = query_class(baseobject=cls)
    return cls


def k8s_custom_object(group: str, version: str, plural: str):
    def wrap(cls):
        if issubclass(cls, models.ClusterModel):
            query_class = pykorm_query.ClusterObjectQuery
        elif issubclass(cls, models.NamespacedModel):
            query_class = pykorm_query.NamespacedObjectQuery
        else:
            raise Exception(f"Class {cls} doesn't seem to inherit from either ClusterModel nor NamespacedModel")

        return _process_cls(cls, query_class, group, version, plural)

    return wrap



class Pykorm:
    def __init__(self):
        # To be changed once https://github.com/kubernetes-client/python/issues/1005 is fixed
        try:
            kubernetes.config.load_kube_config()
        except:  # XXX
            kubernetes.config.load_incluster_config()

    def save(self, obj: models.PykormModel):
        obj.query._save(obj)

    def delete(self, obj: models.PykormModel):
        obj.query._delete(obj)




from setuptools import setup
from setuptools import find_packages
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='pykorm',
    version='0.46.0',
    description='Pykorm: a dead simple Kubernetes ORM',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Frankkkkk/pykorm',
    author='Frank Villaro-Dixon - Infomaniak Network SA',
    author_email='frank.villaro@infomaniak.com',
    license='GNU General Public License v3 (GPLv3)',
    packages=find_packages(),
    zip_safe=False,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires=">=3.7",
    install_requires=['kubernetes'])

